"""Amazon S3 storage connector."""
import json
import logging
import mimetypes
import os
import threading
import uuid
from pathlib import Path
from typing import Dict, Optional

import boto3
import botocore

from .baseconnector import (
    BaseStorageConnector,
    ConnectorType,
    validate_url,
    validate_urls,
)

logger = logging.getLogger(__name__)


class AwsS3Connector(BaseStorageConnector):
    """Amazon S3 storage connector."""

    REQUIRED_SETTINGS = ["bucket"]
    CONNECTOR_TYPE = ConnectorType.S3

    def __init__(self, config: dict, name: str):
        """Connector initialization."""
        super().__init__(config, name)
        self.bucket_name = config["bucket"]
        self.supported_hash = ["awss3etag"]
        self.refresh_hash_after_transfer = ["awss3etag"]
        self.hash_property = {"awss3etag": "ETag"}
        self.multipart_threshold = self.config.get(
            "multipart_threshold", self.CHUNK_SIZE
        )
        self.multipart_chunksize = self.config.get(
            "multipart_chunksize", self.CHUNK_SIZE
        )
        self.use_threads = True

        # Ensured by TLS protocol used for transport.
        self.get_ensures_data_integrity = True
        # Boto library computes MD5 checksums for every chunk of data it
        # transmits to the bucket and compares it to the actual md5 of the
        # data received by the server.
        self.put_ensures_data_integrity = True
        self._init_lock = threading.Lock()
        self._initialized = False

    def duplicate(self):
        """Duplicate existing connector.

        The low-level rountines of boto3 are thread safe. As long as we do not
        use session and resources in threads we should be safe, so no clonning
        is necessary.
        """
        return self

    def _initialize(self):
        """Initializaton."""
        logger.debug("Initializing S3 connector")
        session_kwargs = {"region_name": self.config.get("region_name")}
        # Credentials may be given explicitely in config. When they are not
        # the system credentials are used instead.
        if "credentials" in self.config:
            logger.debug("Reading credentials from %s.", self.config["credentials"])
            with open(self.config["credentials"]) as f:
                settings = json.load(f)
                session_kwargs["aws_access_key_id"] = settings["AccessKeyId"]
                session_kwargs["aws_secret_access_key"] = settings["SecretAccessKey"]

        self.session = boto3.Session(**session_kwargs)
        self.original_session = self.session

        # Assume role if needed.
        if "role_arn" in self.config:
            logger.debug("Assuming role %s.", self.config["role_arn"])
            # This has to be unique identifier and can be seen in logs.
            self._session_name = "s3connector"
            session_credentials = (
                botocore.credentials.RefreshableCredentials.create_from_metadata(
                    metadata=self._refresh_credentials_metadata(),
                    refresh_using=self._refresh_credentials_metadata,
                    method="sts-assume-role",
                )
            )
            botocore_session = botocore.session.get_session()
            botocore_session._credentials = session_credentials
            if "region_name" in self.config:
                botocore_session.set_config_variable(
                    "region", self.config["region_name"]
                )
            # Replace the session with the new one.
            self.session = boto3.Session(botocore_session=botocore_session)

        self.client = self.session.client(
            "s3",
            config=botocore.client.Config(
                signature_version="s3v4", max_pool_connections=50
            ),
        )

    def _refresh_credentials_metadata(self, duration: int = 3600):
        """Prepare metadata for refreshing credentials when assuming role."""
        logger.debug(
            "Creating metadata for refreshable credentials for role '%s', duration %d.",
            self.config["role_arn"],
            duration,
        )
        params = {
            "RoleArn": self.config["role_arn"],
            "RoleSessionName": str(uuid.uuid4()),
            "DurationSeconds": duration,
        }
        response = (
            self.original_session.client("sts").assume_role(**params).get("Credentials")
        )
        return {
            "access_key": response.get("AccessKeyId"),
            "secret_key": response.get("SecretAccessKey"),
            "token": response.get("SessionToken"),
            "expiry_time": response.get("Expiration").isoformat(),
        }

    def __getattr__(self, name):
        """Lazy initialize some attributes."""
        requires_initialization = [
            "client",
            "awss3",
            "sts",
            "session",
            "original_session",
        ]
        if name not in requires_initialization:
            raise AttributeError()
        with self._init_lock:
            if not self._initialized:
                self._initialize()
                self._initialized = True
        return getattr(self, name)

    @validate_url
    def push(self, stream, url, chunk_size=BaseStorageConnector.CHUNK_SIZE, hashes={}):
        """Push data from the stream to the given URL."""
        url = os.fspath(url)
        mime_type = mimetypes.guess_type(url)[0]
        extra_args = {} if mime_type is None else {"ContentType": mime_type}
        extra_args["Metadata"] = {"_upload_chunk_size": str(chunk_size)}
        extra_args["Metadata"].update(hashes)
        self.client.upload_fileobj(
            stream,
            self.bucket_name,
            url,
            Config=self._get_transfer_config(chunk_size),
            ExtraArgs=extra_args,
        )

    def multipart_push_start(self, url, size=None):
        """Start a multipart upload.

        :returns: the upload id.
        """
        url = os.fspath(url)
        mime_type = mimetypes.guess_type(url)[0]
        upload_args = {"Bucket": self.bucket_name, "Key": url}
        if mime_type is not None:
            upload_args["ContentType"] = mime_type
        response = self.client.create_multipart_upload(**upload_args)
        return response["UploadId"]

    def multipart_push(self, upload_id, url, part_number, chunk_size, data, md5=None):
        """Upload a single part of multipart upload."""
        result = self.client.upload_part(
            Body=data,
            Bucket=self.bucket_name,
            ContentMD5=md5,
            Key=os.fspath(url),
            PartNumber=part_number,
            UploadId=upload_id,
        )
        return {"ETag": result["ETag"], "PartNumber": part_number}

    def multipart_push_complete(self, upload_id, url, completed_chunks):
        """Complete the multipart push."""
        return self.client.complete_multipart_upload(
            Bucket=self.bucket_name,
            Key=os.fspath(url),
            UploadId=upload_id,
            MultipartUpload={"Parts": completed_chunks},
        )

    def multipart_push_abort(self, upload_id, url):
        """Abort multiport upload."""
        return self.client.abort_multipart_upload(
            Bucket=self.bucket_name,
            Key=os.fspath(url),
            UploadId=upload_id,
        )

    @validate_urls
    @validate_url
    def delete(self, url, urls):
        """Remove objects."""
        # At most 1000 objects can be deleted at the same time.
        max_chunk = 1000
        for i in range(0, len(urls), max_chunk):
            next_chunk = urls[i : i + max_chunk]
            objects = [
                {"Key": os.fspath(self.base_path / url / delete_url)}
                for delete_url in next_chunk
            ]
            self.client.delete_objects(
                Bucket=self.bucket_name, Delete={"Objects": objects, "Quiet": True}
            )

    @validate_url
    def get(self, url, stream, chunk_size=BaseStorageConnector.CHUNK_SIZE):
        """Get data from the given URL and write it into the given stream."""
        chunk_size = max(chunk_size, self.multipart_threshold)
        self.client.download_fileobj(
            self.bucket_name,
            os.fspath(url),
            stream,
            Config=self._get_transfer_config(chunk_size),
        )

    def _get_transfer_config(self, chunk_size=BaseStorageConnector.CHUNK_SIZE):
        """Get transfer config object."""
        chunk_size = max(chunk_size, self.multipart_threshold)
        return boto3.s3.transfer.TransferConfig(
            multipart_threshold=self.multipart_threshold,
            multipart_chunksize=chunk_size,
            use_threads=self.use_threads,
        )

    @validate_url
    def get_object_list(self, url):
        """Get a list of objects stored bellow the given URL."""
        url = os.path.join(url, "")
        paginator = self.client.get_paginator("list_objects_v2")
        kwargs = {"Bucket": self.bucket_name, "Prefix": url}
        ret = []
        for page in paginator.paginate(**kwargs):
            try:
                contents = page["Contents"]
            except KeyError:
                break
            for obj in contents:
                ret.append(Path(obj["Key"]).relative_to(url).as_posix())
        return ret

    @validate_url
    def get_hash(self, url, hash_type):
        """Get the hash of the given type for the given object."""
        try:
            response = self.client.head_object(
                Bucket=self.bucket_name, Key=os.fspath(url)
            )
            if hash_type in self.hash_property:
                return response[self.hash_property[hash_type]].strip('"')
            else:
                return response["Metadata"].get(hash_type)
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Code"] == "404":
                return None
            else:
                # Something else has gone wrong.
                raise

    @validate_url
    def get_hashes(self, url, hash_types):
        """Get the hash of the given type for the given object."""
        hashes = dict()
        try:
            response = self.client.head_object(
                Bucket=self.bucket_name, Key=os.fspath(url)
            )
            for hash_type in hash_types:
                if hash_type in self.hash_property:
                    hashes[hash_type] = response[self.hash_property[hash_type]].strip(
                        '"'
                    )
                else:
                    hashes[hash_type] = response["Metadata"].get(hash_type)
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Code"] == "404":
                return None
            else:
                # Something else has gone wrong.
                raise
        return hashes

    @validate_url
    def exists(self, url):
        """Get if the object at the given URL exist."""
        try:
            self.client.head_object(Bucket=self.bucket_name, Key=os.fspath(url))
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Code"] == "404":
                return False
            else:
                # Something else has gone wrong.
                raise
        else:
            return True

    @validate_url
    def set_hashes(self, url, hashes):
        """Set the  hashes for the given object."""
        # Changing metadata on existing objects in S3 is annoyingly hard.
        # See
        # https://boto3.amazonaws.com/v1/documentation/api/1.9.42/guide/s3.html
        # under managed copy for example.
        # If one uses copy_object method proposed by some solutions the e_tag
        # value on object can (and will) change. That causes error downloading
        # since hash check fails.
        url = os.fspath(url)
        head = self.client.head_object(Bucket=self.bucket_name, Key=url)
        content_type = head["ResponseMetadata"]["HTTPHeaders"]["content-type"]
        meta = head["Metadata"]
        chunk_size = int(meta["_upload_chunk_size"])
        hashes = {k: v for (k, v) in hashes.items() if k not in self.hash_property}
        meta.update(hashes)
        copy_source = {
            "Bucket": self.bucket_name,
            "Key": url,
        }
        self.client.copy(
            copy_source,
            self.bucket_name,
            url,
            ExtraArgs={
                "Metadata": meta,
                "MetadataDirective": "REPLACE",
                "ContentType": content_type,
            },
            Config=self._get_transfer_config(chunk_size),
        )

    @property
    def base_path(self):
        """Get a base path for this connector."""
        return Path("")

    @validate_url
    def url(self, url):
        """Return the URL to the given file in connector native notation.

        AWS native notation for file is S3 bucket is s3://bucket/file .
        """
        return f"s3://{self.bucket_name}/{url}"

    @validate_url
    def presigned_url(self, url, expiration=60, force_download=False):
        """Create a presigned URL.

        The URL is used to obtain temporary access to the object ar the
        given URL using only returned URL.

        :param expiration: expiration time of the link (in seconds), default
            is one minute.

        :param force_download: force download.

        :returns: URL that can be used to access object or None.
        """
        content_disposition = "attachment" if force_download else "inline"
        content_disposition += f';filename="{Path(url).name}"'
        response = None
        try:
            response = self.client.generate_presigned_url(
                "get_object",
                Params={
                    "Bucket": self.bucket_name,
                    "Key": os.fspath(url),
                    "ResponseContentDisposition": content_disposition,
                },
                ExpiresIn=expiration,
            )
        except botocore.exceptions.ClientError:
            logger.exception("Error creating presigned URL")

        # The response contains the presigned URL
        return response

    def temporary_credentials(
        self, prefix: str, duration: int = 900, role_arn: Optional[str] = None, **kwargs
    ) -> Dict:
        """Get the temporary credentials for role_arn.

        The credentials are further limited to access only objects which names
        start with '/prefix/'.

        :attr role_arn: arn of the role to assume. If none is given it is taken
            from connector settings.

        :attr prefix: the string representing prefix.

        :attr duration: the duration of the credentials in seconds (defaults to 60).

        :raises KeyError: when no role_arn is gives as argument or connector
            setting.

        :returns: the dictionary with credentials, see
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sts.html#STS.Client.assume_role
        for detailed description.
        """
        # Policy to add access only to the objects named '/prefix/*'.
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "AllowAllS3ActionsInUserFolder",
                    "Effect": "Allow",
                    "Action": [
                        "s3:PutObject",
                        "s3:AbortMultipartUpload",
                        "s3:GetObject",
                    ],
                    "Resource": [f"arn:aws:s3:::{self.bucket_name}/{prefix}/*"],
                },
                {
                    "Sid": "AllowUserKMSUsage",
                    "Effect": "Allow",
                    "Action": ["kms:GenerateDataKey", "kms:Decrypt"],
                    "Resource": ["*"],
                },
            ],
        }
        aws_response = self.original_session.client("sts").assume_role(
            RoleArn=role_arn or self.config["role_arn"],
            RoleSessionName="upload-" + prefix,
            DurationSeconds=duration,
            Policy=json.dumps(policy),
        )
        return {
            "credentials": aws_response["Credentials"],
            "bucket_name": self.bucket_name,
            "region": self.config.get("region_name"),
            "prefix": prefix,
        }



"""Entity viewset."""
import re

from drf_spectacular.utils import extend_schema

from django.core.exceptions import ValidationError as DjangoValidationError
from django.db import transaction
from django.db.models import F, Func, OuterRef, Prefetch, Subquery
from django.db.models.functions import Coalesce

from rest_framework import exceptions, serializers, status
from rest_framework.decorators import action, authentication_classes
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from resolwe.flow.filters import EntityFilter
from resolwe.flow.models import AnnotationValue, Data, DescriptorSchema, Entity
from resolwe.flow.models.annotations import AnnotationField, HandleMissingAnnotations
from resolwe.flow.serializers import EntitySerializer
from resolwe.flow.serializers.annotations import AnnotationsSerializer
from resolwe.observers.mixins import ObservableMixin
from resolwe.observers.views import BackgroundTaskSerializer
from resolwe.process.descriptor import ValidationError

from .collection import BaseCollectionViewSet
from .utils import get_collection_for_user


class MoveEntityToCollectionSerializer(serializers.Serializer):
    """Serializer for moving entities to a collection."""

    ids = serializers.ListField(child=serializers.IntegerField())
    destination_collection = serializers.IntegerField()
    handle_missing_annotations = serializers.ChoiceField(
        choices=HandleMissingAnnotations._member_names_,
        default=HandleMissingAnnotations.ADD.name,
    )
    confirm_annotations_remove = serializers.BooleanField(default=False)


class DuplicateEntitySerializer(serializers.Serializer):
    """Serializer for duplicating entities."""

    ids = serializers.ListField(child=serializers.IntegerField())
    inherit_collection = serializers.BooleanField(default=False)


class EntityViewSet(ObservableMixin, BaseCollectionViewSet):
    """API view for entities."""

    serializer_class = EntitySerializer
    filterset_class = EntityFilter
    qs_descriptor_schema = DescriptorSchema.objects.select_related("contributor")

    data_count_subquery = (
        Data.objects.filter(entity_id=OuterRef("pk"))
        .annotate(count=Func(F("id"), function="Count"))
        .values("count")
    )
    data_status_subquery = (
        Data.objects.filter(entity_id=OuterRef("pk"))
        .annotate(
            count=Coalesce(
                Func(
                    F("status"),
                    function="ARRAY_AGG",
                    template="%(function)s(DISTINCT %(expressions)s)",
                ),
                [],
            )
        )
        .values("count")
    )
    queryset = (
        Entity.objects.select_related("contributor", "descriptor_schema__contributor")
        .prefetch_related(
            Prefetch("collection", queryset=BaseCollectionViewSet.queryset)
        )
        .annotate(data_statuses=Subquery(data_status_subquery))
        .annotate(data_count=Subquery(data_count_subquery))
    )

    def order_queryset(self, queryset):
        """Order queryset by annotation values.

        The format of the ordering field is annotations__field_id, possibly prefixed by
        a minus sign. The minus sign indicates descending order.
        """
        if ordering := self.request.query_params.get("ordering"):
            order_by = []
            regex = re.compile(r"-?annotations__(?P<field_id>\d+)")
            fields = [field.strip() for field in ordering.split(",")]
            for match in filter(None, map(regex.match, fields)):
                field_id = match.group("field_id")
                # Always filter by user-visible attribute annotation_label.
                annotation_value = AnnotationValue.objects.filter(
                    entity_id=OuterRef("pk"), field_id=field_id
                ).values("_value__label")
                annotate = {f"_order_{field_id}": Subquery(annotation_value)}
                queryset = queryset.annotate(**annotate)
                sign = "-" if match.string.startswith("-") else ""
                order_by.append(f"{sign}_order_{field_id}")
            if order_by:
                queryset = queryset.order_by(*order_by)
        return queryset

    def filter_queryset(self, queryset):
        """Filter queryset."""
        # First apply the default filters.
        queryset = super().filter_queryset(queryset)
        # Order by annotation values if required.
        return self.order_queryset(queryset)

    def _get_entities(self, user, ids):
        """Return entities queryset based on provided entity ids."""
        queryset = Entity.objects.filter(id__in=ids).filter_for_user(user)
        actual_ids = queryset.values_list("id", flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Entities with the following ids not found: {}".format(
                    ", ".join(map(str, missing_ids))
                )
            )

        return queryset

    @extend_schema(
        request=MoveEntityToCollectionSerializer(), responses={status.HTTP_200_OK: None}
    )
    @action(detail=False, methods=["post"])
    def move_to_collection(self, request, *args, **kwargs):
        """Move samples from source to destination collection.

        The request accepts argument 'handle_missing_annotations'.The possible
        values for it are 'ADD' and 'REMOVE'. When 'REMOVE' is set without
        arguments the 'confirm_annotations_delete' argument must be set to
        True.

        :raises ValidationError: when handle_missing_annotations has incorrect
            value.
        :raises ValidationError: when handle_missing_annotations is set to
            'REMOVE' and 'confirm_annotations_delete' is set.
        """

        serializer = MoveEntityToCollectionSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        ids = serializer.validated_data["ids"]
        dst_collection_id = serializer.validated_data["destination_collection"]
        dst_collection = get_collection_for_user(dst_collection_id, request.user)
        entity_qs = self._get_entities(request.user, ids)

        missing_annotations = {
            entity: AnnotationField.objects.filter(collection=entity.collection)
            .exclude(collection=dst_collection)
            .distinct()
            for entity in entity_qs
        }

        if missing_annotations:
            # How to handle annotation fields present in the current collection
            # but missing in new one. We have two options:
            # - add (default): add the missing annotation fields to the new
            #   collection and keep the values.
            # - delete: remove the missing annotation values from the samples.
            #   This option requires explicit confirmation by 'confirm_remove'
            #   argument or exception with explanation is raised.
            handle_missing_annotations = serializer.validated_data.get(
                "handle_missing_annotations"
            )

            if handle_missing_annotations == HandleMissingAnnotations.REMOVE:
                if not serializer.validated_data["confirm_annotations_remove"]:
                    raise ValidationError(
                        "All annotations not present in the target collection will be "
                        "removed. Set 'confirm_action' argument to 'True' to confirm."
                    )

        entity_qs.move_to_collection(dst_collection, handle_missing_annotations)
        return Response()

    # NOTE: This can be deleted when DRF will support select_for_update
    #       on updates and ResolweUpdateModelMixin will use it.
    #       https://github.com/encode/django-rest-framework/issues/4675
    def update(self, request, *args, **kwargs):
        """Update an entity.

        Original queryset produces a temporary database table whose rows
        cannot be selected for an update. As a workaround, we patch
        get_queryset function to return only Entity objects without
        additional data that is not needed for the update.
        """
        orig_get_queryset = self.get_queryset

        def patched_get_queryset():
            """Patched get_queryset method."""
            entity_ids = orig_get_queryset().values_list("id", flat=True)
            return Entity.objects.filter(id__in=entity_ids)

        self.get_queryset = patched_get_queryset

        resp = super().update(request, *args, **kwargs)

        self.get_queryset = orig_get_queryset
        return resp

    @extend_schema(
        request=DuplicateEntitySerializer(),
        responses={status.HTTP_200_OK: None},
    )
    @action(detail=False, methods=["post"])
    @authentication_classes([IsAuthenticated])
    def duplicate(self, request, *args, **kwargs):
        """Duplicate (make copy of) ``Entity`` models.

        The objects are duplicated in the background and the details of the background
        task handling the duplication are returned.
        """
        serializer = DuplicateEntitySerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        inherit_collection = serializer.validated_data["inherit_collection"]
        ids = serializer.validated_data["ids"]

        queryset = Entity.objects.filter(id__in=ids).filter_for_user(request.user)
        actual_ids = queryset.values_list("id", flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Entities with the following ids not found: {}".format(
                    ", ".join(map(str, missing_ids))
                )
            )

        task = queryset.duplicate(
            contributor=request.user, inherit_collection=inherit_collection
        )
        serializer = BackgroundTaskSerializer(task)
        return Response(serializer.data)

    @extend_schema(
        request=AnnotationsSerializer(many=True), responses={status.HTTP_200_OK: None}
    )
    @action(detail=True, methods=["post"])
    def set_annotations(self, request, pk=None):
        """Add the given list of AnnotaitonFields to the given collection."""
        # No need to check for permissions, since post requires edit by default.
        entity = self.get_object()
        # Read and validate the request data.
        serializer = AnnotationsSerializer(data=request.data, many=True)
        serializer.is_valid(raise_exception=True)
        annotations = [
            (entry["field"], entry["value"]) for entry in serializer.validated_data
        ]
        annotation_fields = [e[0] for e in annotations]
        # The following dict is a mapping from annotation field id to the annotation
        # value id.
        existing_annotations = dict(
            entity.annotations.filter(field__in=annotation_fields).values_list(
                "field_id", "id"
            )
        )

        validation_errors = []
        to_create = []
        to_update = []
        for field, value in annotations:
            annotation_id = existing_annotations.get(field.id)
            append_to = to_create if annotation_id is None else to_update
            annotation = AnnotationValue(
                entity_id=entity.id, field_id=field.id, value=value, id=annotation_id
            )
            try:
                annotation.validate()
            except DjangoValidationError as e:
                validation_errors += e
            append_to.append(annotation)

        if validation_errors:
            raise DjangoValidationError(validation_errors)

        with transaction.atomic():
            AnnotationValue.objects.bulk_create(to_create)
            AnnotationValue.objects.bulk_update(to_update, ["_value"])
        return Response()



""".. Ignore pydocstyle D400.

======================
Resolwe Test Framework
======================

.. automodule:: resolwe.test.testcases

.. automodule:: resolwe.test.utils
   :members:

"""
from resolwe.test.testcases import TestCase, TestCaseHelpers, TransactionTestCase
from resolwe.test.testcases.api import ResolweAPITestCase, TransactionResolweAPITestCase
from resolwe.test.testcases.process import ProcessTestCase
from resolwe.test.utils import (
    check_docker,
    check_installed,
    has_process_tag,
    is_testing,
    tag_process,
    with_custom_executor,
    with_docker_executor,
    with_null_executor,
    with_resolwe_host,
)

__all__ = (
    "TestCase",
    "TestCaseHelpers",
    "TransactionTestCase",
    "ResolweAPITestCase",
    "TransactionResolweAPITestCase",
    "ProcessTestCase",
    "check_docker",
    "check_installed",
    "with_custom_executor",
    "with_docker_executor",
    "with_null_executor",
    "with_resolwe_host",
    "tag_process",
    "has_process_tag",
    "is_testing",
)
